@@ -38,7 +38,7 @@ def downsampled_image_dims_from_desired_num_pixels(image_dims, num_pixels, maxim
 
 
 def create_sampled_pixel_coords_image(image_dims, samples_per_dim, batch_shape=None, normalized=False, randomize=True,
-                                      homogeneous=False, dev_str='cpu'):
+                                      homogeneous=False, dev_str=None):
     """
     Create image of randomly sampled homogeneous integer :math:`xy` pixel co-ordinates :math:`\mathbf{X}\in\mathbb{Z}^{h×w×3}`,
     stored as floating point values. The origin is at the top-left corner of the image, with :math:`+x` rightwards, and
@@ -89,7 +89,8 @@ def create_sampled_pixel_coords_image(image_dims, samples_per_dim, batch_shape=N
         # BS x DH x DW x 2
         rand_offsets = ivy.concatenate((rand_x, rand_y), -1)
         downsam_pix_coords += rand_offsets
-    downsam_pix_coords = ivy.round(downsam_pix_coords)
+    downsam_pix_coords = ivy.clip(ivy.round(downsam_pix_coords), ivy.array([0.]*2, dev_str=dev_str),
+                                  ivy.array(list(reversed(image_dims)), dtype_str='float32', dev_str=dev_str)-1)
 
     if normalized:
         downsam_pix_coords /=\
@@ -103,7 +104,7 @@ def create_sampled_pixel_coords_image(image_dims, samples_per_dim, batch_shape=N
     return downsam_pix_coords
 
 
-def sample_images(list_of_images, num_pixels, batch_shape, image_dims, dev_str='cpu'):
+def sample_images(list_of_images, num_pixels, batch_shape, image_dims, dev_str=None):
     """
     Samples each image in a list of aligned images at num_pixels random pixel co-ordinates, within a unifrom grid
     over the image.
@@ -169,23 +170,6 @@ def sample_images(list_of_images, num_pixels, batch_shape, image_dims, dev_str='
     return ivy.split(combined_img_sampled, image_channels, -1)
 
 
-def sinusoid_positional_encoding(x, embedding_length=10):
-    """
-    Perform sinusoid positional encoding of the inputs.
-
-    :param x: input array to encode *[batch_shape, dim]*
-    :type x: array
-    :param embedding_length: Length of the embedding. Default is 10.
-    :type embedding_length: int, optional
-    :return: The new positionally encoded array *[batch_shape, dim+dim*2*embedding_length]*
-    """
-    rets = [x]
-    for i in range(embedding_length):
-        for fn in [ivy.sin, ivy.cos]:
-            rets.append(fn(2.**i * x))
-    return ivy.concatenate(rets, -1)
-
-
 # noinspection PyUnresolvedReferences
 def sampled_volume_density_to_occupancy_probability(density, inter_sample_distance):
     """
@@ -279,28 +263,35 @@ def render_rays_via_termination_probabilities(ray_term_probs, features, render_v
     return rendering, var
 
 
-def render_implicit_features_and_depth_from_net_inputs(
-        network_fn, query_points, samples_per_ray, z_vals, input_features=None, render_depth=True, render_feats=True,
-        render_variance=False, with_grads=True, v=None):
+def render_implicit_features_and_depth(network_fn, rays_o, rays_d, near, far, samples_per_ray, timestamps=None,
+                                       render_depth=True, render_feats=True, render_variance=False, inter_feat_fn=None,
+                                       with_grads=True, v=None):
     """
     Render an rgb-d image, given an implicit rgb and density function conditioned on xyz data.
 
     :param network_fn: the implicit function.
     :type network_fn: callable
-    :param query_points: The points at which to query the implicit network.
-    :type query_points: array
-    :param input_features: The features passed as input to the implicit function. Default is None.
-    :type input_features: array, optional
+    :param rays_o: The camera center *[batch_shape,3]*
+    :type rays_o: array
+    :param rays_d: The rays in world space *[batch_shape,ray_batch_shape,3]*
+    :type rays_d: array
+    :param near: The near clipping plane values *[batch_shape,ray_batch_shape]*
+    :type near: array
+    :param far: The far clipping plane values *[batch_shape,ray_batch_shape]*
+    :type far: array
     :param samples_per_ray: The number of stratified samples to use along each ray
     :type samples_per_ray: int
-    :param z_vals: The depth values associated with each of the sampled points.
-    :type z_vals: array
+    :param timestamps: The timestamps associated with each image. Default is None. *[batch_shape,1]*
+    :type timestamps: array, optional
     :param render_depth: Whether to render the depth. Default is True.
     :type render_depth: bool, optional
     :param render_feats: Whether to render the features. Default is True.
     :type render_feats: bool, optional
     :param render_variance: Whether to also render the feature variance. Default is False.
     :type render_variance: bool, optional
+    :type render_variance: bool, optional
+    :param inter_feat_fn: Function to extract interpolated features from world-coords *[batch_shape,ray_batch_shape,3]*
+    :type inter_feat_fn: callable, optional
     :param with_grads: Whether to track gradients during the network forward pass. Defualt is True.
     :type with_grads: bool, optional
     :param v: The container of trainable variables for the implicit model. default is to use internal variables.
@@ -309,26 +300,47 @@ def render_implicit_features_and_depth_from_net_inputs(
     """
 
     # shapes
-    batch_shape = list(query_points.shape[:-3])
+    batch_shape = list(rays_o.shape[:-1])
     num_batch_dims = len(batch_shape)
-    ray_batch_shape = list(query_points.shape[num_batch_dims:-2])
+    ray_batch_shape = list(rays_d.shape[num_batch_dims:-1])
+    flat_ray_batch_size = int(np.prod(ray_batch_shape))
     num_ray_batch_dims = len(ray_batch_shape)
     total_batch_shape = batch_shape + ray_batch_shape
-    flat_total_batch_size = np.prod(total_batch_shape)
+
+    # Compute 3D query points
+
+    # BS x RBS x SPR x 1
+    z_vals = ivy.expand_dims(stratified_sample(near, far, samples_per_ray), -1)
+
+    # BS x RBS x 1 x 3
+    rays_d = ivy.expand_dims(rays_d, -2)
+    rays_o = ivy.broadcast_to(ivy.reshape(rays_o, batch_shape + [1]*(num_ray_batch_dims+1) + [3]), rays_d.shape)
+
+    # BS x RBS x SPR x 3
+    pts = rays_o + rays_d * z_vals
 
     # flatten
 
-    # (BSxRBSxSPR) x 3
-    pts_flat = ivy.reshape(query_points, (flat_total_batch_size*samples_per_ray, 3))
+    # BS x FRBS x SPR x 3
+    pts_flat = ivy.reshape(pts, batch_shape + [flat_ray_batch_size, samples_per_ray, 3])
 
-    # (BSxRBSxSPR) x IF
-    if input_features is not None:
-        input_features = ivy.reshape(input_features, (flat_total_batch_size*samples_per_ray, -1))
+    # input features
+
+    if inter_feat_fn is not None:
+
+        # BS x RBS x SPR x F
+        features = inter_feat_fn(pts)
+
+        # BS x FRBS x SPR x F
+        features_flat = ivy.reshape(features, batch_shape + [flat_ray_batch_size, samples_per_ray, -1])
+
+    else:
+        features_flat = None
 
     # Run network
 
     # BSPQ x OF,    BSPQ
-    feat, densities = network_fn(query_points, input_features, with_grads=with_grads, v=v)
+    feat, densities = network_fn(pts_flat, features_flat, timestamps, with_grads=with_grads, v=v)
 
     # BS x RBS x SPR
     densities = ivy.reshape(densities, total_batch_shape + [samples_per_ray])
@@ -365,102 +377,4 @@ def render_implicit_features_and_depth_from_net_inputs(
             ret_vals.append(radial_depth)
 
     # up to BS x RBS x OF, BS x RBS x OF, BS x RBS x 1, BS x RBS x 1
-    return ret_vals
-
-
-def render_implicit_features_and_depth(network_fn, rays_o, rays_d, near, far, samples_per_ray, render_depth=True,
-                                       render_feats=True, render_variance=False, batch_size_per_query=512 * 64,
-                                       inter_feat_fn=None, with_grads=True, v=None):
-    """
-    Render an rgb-d image, given an implicit rgb and density function conditioned on xyz data.
-
-    :param network_fn: the implicit function.
-    :type network_fn: callable
-    :param rays_o: The camera center *[batch_shape,3]*
-    :type rays_o: array
-    :param rays_d: The rays in world space *[batch_shape,ray_batch_shape,3]*
-    :type rays_d: array
-    :param near: The near clipping plane values *[batch_shape,ray_batch_shape]*
-    :type near: array
-    :param far: The far clipping plane values *[batch_shape,ray_batch_shape]*
-    :type far: array
-    :param samples_per_ray: The number of stratified samples to use along each ray
-    :type samples_per_ray: int
-    :param render_depth: Whether to render the depth. Default is True.
-    :type render_depth: bool, optional
-    :param render_feats: Whether to render the features. Default is True.
-    :type render_feats: bool, optional
-    :param render_variance: Whether to also render the feature variance. Default is False.
-    :type render_variance: bool, optional
-    :param batch_size_per_query: The maximum batch size for querying the implicit network. Default is 1024.
-    :type batch_size_per_query: int, optional
-    :param inter_feat_fn: Function to extract interpolated features from world-coords *[batch_shape,ray_batch_shape,3]*
-    :type inter_feat_fn: callable, optional
-    :param with_grads: Whether to track gradients during the network forward pass. Defualt is True.
-    :type with_grads: bool, optional
-    :param v: The container of trainable variables for the implicit model. default is to use internal variables.
-    :type v: ivy Container of variables
-    :return: The rendered feature *[batch_shape,ray_batch_shape,feat]* and radial depth *[batch_shape,ray_batch_shape,1]*
-    """
-
-    # shapes
-    batch_shape = list(rays_o.shape[:-1])
-    flat_batch_size = np.prod(batch_shape)
-    num_batch_dims = len(batch_shape)
-    ray_batch_shape = list(rays_d.shape[num_batch_dims:-1])
-    flat_ray_batch_size = np.prod(ray_batch_shape)
-    num_ray_batch_dims = len(ray_batch_shape)
-    total_batch_shape = batch_shape + ray_batch_shape
-    flat_total_batch_size = np.prod(total_batch_shape)
-    num_sections = math.ceil(flat_total_batch_size*samples_per_ray/batch_size_per_query)
-
-    # Compute 3D query points
-
-    # BS x RBS x SPR x 1
-    z_vals = ivy.expand_dims(stratified_sample(near, far, samples_per_ray), -1)
-
-    # BS x RBS x 1 x 3
-    rays_d = ivy.expand_dims(rays_d, -2)
-    rays_o = ivy.broadcast_to(ivy.reshape(rays_o, batch_shape + [1]*(num_ray_batch_dims+1) + [3]), rays_d.shape)
-
-    # BS x RBS x SPR x 3
-    pts = rays_o + rays_d * z_vals
-
-    # flatten
-
-    # BS x FRBS x SPR x 1
-    z_vals_flat = ivy.reshape(z_vals, batch_shape + [flat_ray_batch_size, samples_per_ray, 1])
-
-    # BS x FRBS x SPR x 3
-    pts_flat = ivy.reshape(pts, batch_shape + [flat_ray_batch_size, samples_per_ray, 3])
-
-    # prepare for batched render
-
-    network_inputs = [z_vals_flat, pts_flat]
-    if inter_feat_fn is not None:
-
-        # BS x RBS x SPR x F
-        features = inter_feat_fn(pts)
-
-        # BS x FRBS x SPR x F
-        features_flat = ivy.reshape(features, batch_shape + [flat_ray_batch_size, samples_per_ray, -1])
-
-        network_inputs.append(features_flat)
-        func = lambda z, pt, f: render_implicit_features_and_depth_from_net_inputs(
-            network_fn, pt, samples_per_ray, z, f, render_depth, render_feats, render_variance,
-            with_grads=with_grads, v=v)
-    else:
-        func = lambda z, pt: render_implicit_features_and_depth_from_net_inputs(
-            network_fn, pt, samples_per_ray, z, render_depth=render_depth, render_feats=render_feats,
-            render_variance=render_variance, with_grads=with_grads, v=v)
-
-    # batched render
-
-    # up to BS x FRBS x OF, BS x FRBS x OF, BS x FRBS x 1, BS x FRBS x 1
-    rets_flat = ivy.split_func_call(func, network_inputs, int(round(batch_size_per_query / flat_batch_size)), -3, -2)
-
-    if num_ray_batch_dims > 1:
-        # up to BS x RBS x OF, BS x RBS x OF, BS x RBS x 1, BS x RBS x 1
-        return [ivy.reshape(ret, batch_shape + ray_batch_shape + [-1]) for ret in rets_flat]
-    # up to BS x RBS x OF, BS x RBS x OF, BS x RBS x 1, BS x RBS x 1
-    return rets_flat
+    return tuple(ret_vals)
